""" A UI for browsing a plugin. """


# Enthought library imports.
from envisage.api import ExtensionPoint, IApplication, IExtensionPoint, IPlugin
from traits.api import Delegate, HasTraits, Instance, List, Property
from traits.api import Code, Str
from traitsui.api import Item, TableEditor, View, VGroup
from traitsui.table_column import ObjectColumn # fixme: non-api!

class ExtensionPointModel(HasTraits):
    """ A model for browsing an extension point. """

    # The plugin that offered the extension point.
    plugin = Instance(IPlugin)

    # The extension point.
    extension_point = Instance(IExtensionPoint)

    #### 'ExtensionPointModel' interface ######################################



class ExtensionModel(HasTraits):
    """ A model for browsing a contribution to an extension point. """

    # The plugin that made the contribution.
    plugin = Instance(IPlugin)

    #### 'ContributionModel' interface ########################################

    # The Id of the extension point that the contribution is for.
    extension_point_id = Str

    # The contributions to the extension point.
    contributions = List

    ###########################################################################
    # 'ApplicationModel' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _contributions_default(self):
        """ Trait initializer. """

        return plugin.application.get_extensions(self.extension_point_id)


class PluginModel(HasTraits):
    """ A model for browsing a plugin.  """

    # The plugin that we are a model for.
    plugin = Instance(IPlugin)

    #### 'PluginModel' interface ##############################################

    # Models for each of the plugin's extension points.
    extension_point_models = List(ExtensionPointModel)

    # Models for each of the plugin's contributions to extension points.
    extension_models = List(ExtensionModel)

    ###########################################################################
    # 'ApplicationModel' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _extension_models_default(self):
        """ Trait initializer. """

        extension_models = [
            ExtensionModel(
                plugin             = plugin,
                extension_point_id = extension_point.id
            )

            for extension_point in plugin.get_extension_points()
        ]

        return extension_models

    def _extension_point_models_default(self):
        """ Trait initializer. """

        extension_point_models = [
            ExtensionPointModel(
                plugin          = plugin,
                extension_point = extension_point
            )

            for extension_point in plugin.get_extension_points()
        ]

        return extension_point_models


class ApplicationModel(HasTraits):
    """ A model for browsing an application. """

    # The application that we are a model for.
    application = Instance(IApplication)

    #### 'ApplicationModel' interface #########################################

    # Models for each of the application's plugins.
    plugin_models = List(PluginModel)

    ###########################################################################
    # 'ApplicationModel' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _plugin_models_default(self):
        """ Trait initializer. """

        return [PluginModel(plugin=plugin) for plugin in self.application]



extension_point_table_editor = TableEditor(
    columns = [
        ObjectColumn(name='id'),
        #ObjectColumn(name='desc')
    ],

#   selected = 'extension_point_selected',
    editable  = True,
    edit_view = View(Item(name='desc', show_label=False), style='custom')
)

plugin_browser_view = View(
    VGroup(
        Item(name='id'),
        Item(name='name'),
        label='General'
    ),

    VGroup(
        Item(
            name       = 'extension_points',
            editor     = extension_point_table_editor,
            show_label = False
        ),

        label='Extension Points',
    ),

    width  = .8,
    height = .6
)


class ExtensionPointBrowser(HasTraits):
    """ The model used to view extension points.

    This browser is required because 'ExatenionPoint' instances are trait
    *types* and therefore do not have traits themselves and so this class is
    really just a 'traitified' wrapper.

    """

    #### 'ExtensionPointBrowser' interface ####################################

    # The extension point that we are browsing.
    extension_point = Instance(ExtensionPoint)

    # The extension point's globally unique Id.
    id = Str

    # The extension point's description.
    desc = Code#Str

    ###########################################################################
    # 'ExtensionPoint' browser interface.
    ###########################################################################

    def _id_default(self):
        """ Trait initializer. """

        return self.extension_point.id

    def _desc_default(self):
        """ Trait initializer. """

        desc = self.extension_point.desc.strip()

        lines = [line.replace('    ', '', 2) for line in desc.splitlines()]

        return '\n'.join(lines)


# Convenience trait to delegate an attribute to a plugin.
DelegatedToPlugin = Delegate('plugin', modify=True)


class PluginBrowser(HasTraits):
    """ A plugin browser.

    Actually, this class exists just because to use a trait editor we have
    to have a trait to edit!

    """

    #### 'IPlugin' interface ##################################################

    # The plugins Id.
    id = DelegatedToPlugin

    # The plugin name.
    name = DelegatedToPlugin

    #### 'PluginBrowser' interface ############################################

    # The extension points offered by the plugin.
    extension_points = Property(List)

    # The plugin that we are browsing.
    plugin = Instance(IPlugin)

    # The default traits UI view.
    traits_view = plugin_browser_view

    ###########################################################################
    # 'PluginBrowser' interface.
    ###########################################################################

    def _get_extension_points(self):
        """ Property getter. """

        extension_points = [
            ExtensionPointBrowser(extension_point=extension_point)

            for extension_point in self.plugin.extension_points
        ]

        return extension_points


def browse_plugin(plugin):
    """ Browse a plugin. """

    import inspect

    if inspect.isclass(plugin):
        plugin = plugin()

    plugin_browser = PluginBrowser(plugin=plugin)
    plugin_browser.configure_traits(view=plugin_browser_view)

    return

#### EOF ######################################################################



"""
"""

# System library imports
from Pymacs import lisp

# ETS imports
from envisage.plugins.remote_editor.plugins.editor_plugin import \
        EditorPlugin


client = EditorPlugin()
client.register()

def run_text():
    start = lisp.point()
    end = lisp.mark(True)
    if start > end:
        start, end = end, start
    text = lisp.buffer_substring(start, end)
    if len(text):
        client.run_text(text)
    else:
        # TODO Complain in message bar
        pass

def run_file():
    path = lisp.buffer_file_name()
    if path is None:
        # TODO Complain in message bar
        pass
    else:
        client.run_file(path)



""" This code has been inspired from the IPython repository

https://github.com/ipython/ipython/blob/2.x/examples/Embedding/internal_ipkernel.py

"""
from distutils.version import StrictVersion as Version
import sys

import ipykernel
from ipykernel.connect import connect_qtconsole
from ipykernel.kernelapp import IPKernelApp
from tornado import ioloop

from traits.api import Any, HasStrictTraits, Instance, List

NEEDS_IOLOOP_PATCH = Version(ipykernel.__version__) >= Version('4.7.0')


def gui_kernel(gui_backend):
    """ Launch and return an IPython kernel GUI with support.

    Parameters
    ----------
    gui_backend -- string or None
      The GUI mode used to initialize the GUI mode. For options, see
      the `ipython --gui` help pages. If None, the kernel is initialized
      without GUI support.
    """

    kernel = IPKernelApp.instance()

    argv = ['python']
    if gui_backend is not None:
        argv.append('--gui={}'.format(gui_backend))
    kernel.initialize(argv)

    return kernel


class InternalIPKernel(HasStrictTraits):
    """ Represents an IPython kernel and the consoles attached to it.
    """

    #: The IPython kernel.
    ipkernel = Instance(IPKernelApp)

    #: A list of connected Qt consoles.
    consoles = List()

    #: The kernel namespace.
    #: Use `Any` instead of `Dict` because this is an IPython dictionary
    #: object.
    namespace = Any()

    #: The values used to initialize the kernel namespace.
    #: This is a list of tuples (name, value).
    initial_namespace = List()

    #: sys.stdout value at time kernel was started
    _original_stdout = Any()

    #: sys.stderr value at time kernel was started
    _original_stderr = Any()

    def init_ipkernel(self, gui_backend):
        """ Initialize the IPython kernel.

        Parameters
        ----------
        gui_backend -- string
          The GUI mode used to initialize the GUI mode. For options, see
          the `ipython --gui` help pages.
        """
        # The IPython kernel modifies sys.stdout and sys.stderr when started,
        # and doesn't currently provide a way to restore them. So we restore
        # them ourselves at shutdown.
        self._original_stdout = sys.stdout
        self._original_stderr = sys.stderr

        # Start IPython kernel with GUI event loop support
        self.ipkernel = gui_kernel(gui_backend)

        # Since ipykernel 4.7, the io_loop attribute of the kernel is not
        # initialized anymore
        # Reference: https://github.com/enthought/envisage/issues/107
        # Workaround: Retrieve the kernel on the IPykernelApp and set the
        # io_loop without starting it!
        if NEEDS_IOLOOP_PATCH and not hasattr(self.ipkernel.kernel, 'io_loop'):
            self.ipkernel.kernel.io_loop = ioloop.IOLoop.instance()

        # This application will also act on the shell user namespace
        self.namespace = self.ipkernel.shell.user_ns
        self.namespace.update(dict(self.initial_namespace))

    def new_qt_console(self):
        """ Start a new qtconsole connected to our kernel. """
        console = connect_qtconsole(self.ipkernel.connection_file,
                                    argv=['--no-confirm-exit'])
        self.consoles.append(console)
        return console

    def cleanup_consoles(self):
        """ Kill all existing consoles. """
        for c in self.consoles:
            c.kill()
            c.stdout.close()
            c.stderr.close()
        self.consoles = []

    def shutdown(self):
        """ Shutdown the kernel.

        Existing IPython consoles are killed first.
        """
        if self.ipkernel is not None:
            self.cleanup_consoles()
            self.ipkernel.shell.exit_now = True
            self.ipkernel.cleanup_connection_file()

            # The stdout and stderr streams created by the kernel use the
            # IOPubThread, so we need to close them and restore the originals
            # before we shut down the thread. Without this, we get obscure
            # errors of the form "TypeError: heap argument must be a list".
            kernel_stdout = sys.stdout
            if kernel_stdout is not self._original_stdout:
                sys.stdout = self._original_stdout
                kernel_stdout.close()

            kernel_stderr = sys.stderr
            if kernel_stderr is not self._original_stderr:
                sys.stderr = self._original_stderr
                kernel_stderr.close()

            self.ipkernel.iopub_thread.stop()
            self.ipkernel = None
